//----------------------------------------------------------------------------
/// @file indirect_sort.hpp
/// @brief Indirect algorithm
///
/// @author Copyright (c) 2010 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __COUNTERTREE_UTIL_SORT_INDIRECT_SORT_HPP
#define __COUNTERTREE_UTIL_SORT_INDIRECT_SORT_HPP

#include <boost/countertree/util/algorithm.hpp>
#include <boost/countertree/util/atomic.hpp>
#include <vector>
#include <type_traits>

namespace boost
{
namespace countertree
{
namespace util
{
namespace sort
{
using namespace boost::countertree::util;
using boost::countertree::util::MS1B;

template <class iter_t>
void create_index (iter_t first, iter_t last, std::vector<iter_t> &VP )
{   //----------------------------- begin -------------------------------
    auto N1 = last-first ;
    assert ( N1 >= 0 );
    size_t N = (size_t ) N1 ;
    VP.clear() ;
    VP.resize ( N, first);
    iter_t Alfa = first ;
    for ( size_t i =0 ; i <  N ; ++i, ++Alfa) VP[i] = Alfa ;
};
template <class iter_t>
void sort_index (iter_t first, std::vector<iter_t> &VP )
{   //-------------------------- begin -------------------------------------
    typedef typename iter_value<iter_t>::type       value_t ;
    size_t A  = 0 , B =0 , Pos =0 , N = VP.size();
    iter_t itA, itB ;
    while ( Pos < N )
    //{   for ( ; Pos < N and get_pos( first, VP[Pos]) == Pos ; ++Pos);
    {   for ( ; Pos < N and (size_t (VP[Pos]-first)) == Pos ; ++Pos);
        if ( Pos == N ) return ;
        A = B = Pos ;
        itA = first + A  ;
        value_t Aux = std::move ( *itA);
        while ( (B = (size_t (VP[A]-first)))!= Pos  )
        {   VP[A] = itA;
            itB = first + B ;
            *itA = std::move ( *itB);
            itA = itB ;
            A = B ;
        };
        *itA = std::move ( Aux ) ;
        VP[A] = itA ;
        ++Pos ;
    };
};



//-----------------------------------------------------------------------------
//  function : indirect_sort
/// @brief
/// @param [in/out]
/// @exception
/// @return
/// @remarks
//-----------------------------------------------------------------------------
template <  template <class, class> class tag,
            typename iter_t ,
            typename compare = std::less<typename iter_value<iter_t>::type >
        >
struct indirect_sort_method
{   //------------------------ definitions ---------------------------------
    typedef typename iter_value<iter_t>::type       value_t ;
    typedef typename std::vector<iter_t>::iterator  it_ptr ;
    typedef less_ptr_no_null <iter_t, compare>      compare_ptr ;

indirect_sort_method( iter_t first, iter_t last, compare comp= compare()   )
{   //-------------------------- begin -------------------------------------
    std::vector<iter_t> VP ;
    create_index ( first , last , VP);
    tag< it_ptr,compare_ptr> ( VP.begin() , VP.end(), compare_ptr(comp) );
    sort_index ( first , VP)    ;
};

//****************************************************************************
};//    End struct indirect_sort
//****************************************************************************
//-----------------------------------------------------------------------------
//  function : indirect_sort
/// @brief
/// @param [in/out]
/// @exception
/// @return
/// @remarks
//-----------------------------------------------------------------------------
template <  template <class, class> class tag,
            typename iter_t ,
            typename compare = std::less<typename iter_value<iter_t>::type >
        >
struct parallel_indirect_sort_method
{   //------------------------ definitions ---------------------------------
    typedef typename iter_value<iter_t>::type       value_t ;
    typedef typename std::vector<iter_t>::iterator  it_ptr ;
    typedef less_ptr_no_null <iter_t, compare>      compare_ptr ;

parallel_indirect_sort_method( iter_t first, iter_t last, compare comp , NThread NT = NThread()   )
{   //-------------------------- begin -------------------------------------
    std::vector<iter_t> VP ;
    create_index ( first , last , VP);
    tag< it_ptr,compare_ptr> ( VP.begin() , VP.end(), compare_ptr(comp), NT );
    sort_index ( first , VP)    ;
};
parallel_indirect_sort_method( iter_t first, iter_t last, NThread NT = NThread()    )
{   //-------------------------- begin -------------------------------------
    std::vector<iter_t> VP ;
    create_index ( first , last , VP);
    tag< it_ptr,compare_ptr> ( VP.begin() , VP.end(), NT );
    sort_index ( first , VP)    ;
};

//****************************************************************************
};//    End struct indirect_sort
//****************************************************************************
//
//****************************************************************************
};//    End namespace sort
};//    End namespace util
};//    End namespace countertree
};//    End namespace boost
//****************************************************************************
//
#endif
