//----------------------------------------------------------------------------
/// @file benchmark_sort_03.cpp
/// @brief Benchmark of several sort methods with different uint32_t numbers
///
/// @author Copyright (c) 2010 2015 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
//#define __DEBUG_MODE 1

#include <parallel/basic_iterator.h>
#include <parallel/features.h>
#include <parallel/parallel.h>
#include <parallel/algorithm>
#include <omp.h>
#include <stdlib.h>
#include <iostream>
#include <sys/time.h>
#include <sys/resource.h>
#include <boost/countertree/util_sort.hpp>
#include <boost/countertree/util/time_measure.hpp>
#include <vector>
#include "tbb/tbb_stddef.h"
#include "tbb/task_scheduler_init.h"
#include "tbb/parallel_sort.h"



#define NELEM 200000000
using namespace std ;
//using namespace cntree ;

using boost::countertree::NThread ;
using boost::countertree::util::time_point ;
using boost::countertree::util::now;
using boost::countertree::util::subtract_time ;
using namespace __gnu_parallel;
//using namespace __gnu_parallel::parallel_tag;


int Prueba ( void) ;
vector<uint32_t> A  ;

int main ()
{   //--------------------------- begin --------------------------
    cout<<"max. OpenMP threads:"<<omp_get_max_threads();
    cout<<" (processors)"<< omp_get_num_procs()<<endl;
	A.reserve ( NELEM);
    cout<<"Sort of "<< NELEM<<" elements in a vector -------------------\n\n" ;

    cout<<"Sorted elements -----------------------------------------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < NELEM ; i ++)
        A.push_back( NELEM +i) ;
    Prueba() ;

    cout<<"Reverse sorted elements ---------------------------------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < NELEM ; i ++)
        A.push_back( (NELEM<<1) -i) ;
    Prueba() ;

    cout<<"Random elements, few repeated ( rand() )-----------------------\n";
    A.clear();
    for ( uint32_t i = 0 ; i < NELEM  ; i ++)
        A.push_back(rand()) ;
    Prueba() ;

    cout<<"Random elements, quite repeated ( rand()  % (NELEM/2) )---------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < NELEM  ; i ++)
        A.push_back(rand()  % (NELEM/2));
    Prueba() ;

    cout<<"Random element many repeated (rand() % 10000)--------------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < NELEM  ; i ++)
        A.push_back(rand() % 10000);
    Prueba() ;

    cout<<"Equal elements --------------------------------------------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < NELEM  ; i ++) A.push_back( NELEM);
    Prueba() ;
    return 0 ;
};

int Prueba ( void)
{   //---------------------------- begin --------------------------------
	std::vector<uint32_t>  B ;
	B.reserve ( NELEM);
	double duracion ;
	time_point start, finish;

	//cout<<"--------------------- STL std::sort --------------------------\n";
	B = A ;
	start= now() ;
	std::sort (&B[0] , &B[NELEM]  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"STL std::sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};

	//cout<<"---------------- boost::countertree::intro_sort --------------\n";
	B = A ;
	start= now() ;
	boost::countertree::intro_sort (&B[0] , &B[NELEM] );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"countertree::intro_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};

	//cout<<"--------------------- Parallel std::sort ---------------------\n";
	B = A ;
	start= now() ;
	__gnu_parallel::sort (&B[0] , &B[NELEM]);
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
	cout<<"parallel::sort :"<<duracion<<" secs "<<std::endl;
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};


	//cout<<"---------------------tbb::parallel_sort ----------------------\n";
	B = A ;
	start= now() ;
	tbb::parallel_sort (&B[0] , &B[NELEM]  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"tbb::parallel_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};

	//cout<<"--------------- boost::countertree::parallel_sort ------------\n";
	B = A ;
	start= now() ;
	boost::countertree::parallel_sort (&B[0] , &B[NELEM] ,NThread(4) );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"countertree::parallel_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};

	//cout<<"------------------ STL std::stable_sort ---------------------\n";
	cout<<"\n";
	B = A ;
	start= now() ;
	std::stable_sort (&B[0] , &B[NELEM]  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"std::stable_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};


	//cout<<"--------------------- merge_sort ----------------------------\n";
	B = A ;
	start= now() ;
	boost::countertree::merge_sort (&B[0] , &B[NELEM]  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"countertree::merge_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};
	//cout<<"--------------------- Parallel std::sort --------------------\n";
	B = A ;
	start= now() ;
	__gnu_parallel::stable_sort (&B[0] , &B[NELEM]  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
	cout<<"parallel::stable_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};


	//cout<<"----------------- parallel_merge_sort ------------------------\n";
	B = A ;
	start= now() ;
	boost::countertree::parallel_merge_sort (&B[0] , &B[NELEM] ,NThread(4) );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"countertree::parallel_merge_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (NELEM -1) ; i ++ )
	{	if (  B[i+1] < B[i]  ) cout<<"Error in pos :"<<i ;
	};
    cout<<endl;
    return 0 ;
};
