//----------------------------------------------------------------------------
/// @file benchmark_sort_indirect.cpp
/// @brief Benchmark of several sort methods with different sets of data, and
///        the indirect version of the same method
///
/// @author Copyright (c) 2010 2015 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
//#define __DEBUG_MODE 1

#include <parallel/basic_iterator.h>
#include <parallel/features.h>
#include <parallel/parallel.h>
#include <parallel/algorithm>
#include <omp.h>

#include <boost/countertree/util_sort.hpp>
#include <boost/countertree/util/time_measure.hpp>
#include <vector>
#include <iostream>

#define NELEM 100000000


using namespace __gnu_parallel;
using namespace boost::countertree::util::sort ;
using boost::countertree::less_ptr_no_null;

template <class RBRC> void Prueba  ( uint64_t N);

template <uint32_t NN>
struct reburcio
{   uint64_t M[NN];

    reburcio ( const reburcio &R ) { M[0] = R.M[0] ; };
    reburcio ( uint64_t K =0 ) { M[0] = K ; };
    //reburcio & operator = ( const reburcio &R)
    //{   M[0] = R.M[0];
    //    return *this ;
    //};
    reburcio & operator = ( uint64_t K)
    {   M[0] = K;
        return *this ;
    };
    bool operator < ( const reburcio &R) const{ return (M[0]<R.M[0]);};
};


int main ()
{   //---------------------- begin --------------------------------
    Prueba< reburcio<1> > ( NELEM);
    Prueba< reburcio<2> > ( NELEM>>1);
    Prueba< reburcio<4> > ( NELEM>>2);
    Prueba< reburcio<6> > ( NELEM /6);
    Prueba< reburcio<8> > ( NELEM>>3);
    Prueba< reburcio<16> > ( NELEM>>4);
    Prueba< reburcio<32> > ( NELEM>>5);
}

template <class RBRC>
void Prueba  ( uint64_t N)
{   //--------------------------- Inicio --------------------------
    double duracion ;
	time_point start, finish;
	NThread  NT ;

    std::cout<<"Size of element "<<sizeof(RBRC);
    std::cout<<" Number of elements :"<<N<<std::endl;
    std::vector <RBRC> V, V2 ;
    V.reserve ( N) ;
    for ( uint64_t i  =0 ; i < N ; ++i)
        V2.push_back( rand());

    //------------------------------- parallel_sort -------------------------
    V = V2 ;
    start = now () ;
    parallel_sort(V.begin(),V.end() ,NT);
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    std::cout<<"parallel_sort :"<<duracion<<" sec\n";

    //------------------------------- indirect_parallel_sort -------------------------
    V = V2 ;
    start = now () ;
    indirect_parallel_sort(V.begin(),V.end() ,NT);
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    std::cout<<"indirect_parallel_sort :"<<duracion<<" sec\n";

    //------------------------------- parallel_stable_sort ------------------
    V = V2 ;
    start = now () ;
    parallel_stable_sort(V.begin(),V.end() ,NT);
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    std::cout<<"parallel_stable_sort :"<<duracion<<" sec\n";

    //------------------------------- indirect_parallel_stable_sort ------------------
    V = V2 ;
    start = now () ;
    indirect_parallel_stable_sort(V.begin(),V.end() ,NT);
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    std::cout<<"indirect_parallel_stable_sort :"<<duracion<<" sec\n";
    V.clear() ;




/*
    //--------------------------------------------------------------------
    // Indirect sorting
    //--------------------------------------------------------------------
    std::vector<RBRC*> VP  ( N , nullptr);
    for ( uint64_t i =0 ; i < N ; ++i)
        VP[i] = & V2[i];

    //----------------- Indirect parallel_sort ------------------------------
    start = now () ;
    parallel_sort(VP.begin(),VP.end(),less_ptr_no_null<RBRC>(), NT );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    std::cout<<"Indirect parallel_sort - Tiempo empleado :"<<duracion<<" segundos\n";

    //---------------------- Indirect parallel_stable_sort ---------------------
    for ( uint64_t i =0 ; i < N ; ++i)
        VP[i] = & V2[i];
    start = now () ;
    //__gnu_parallel::stable_sort (VP.begin(),VP.end(),less_ptr_no_null<RBRC>()  );
    parallel_stable_sort(VP.begin(),VP.end(),less_ptr_no_null<RBRC>(), NT );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    std::cout<<"Indirect parallel_stable_sort - Tiempo empleado :"<<duracion<<" segundos\n";
    VP.clear () ;
*/
    std::cout<<std::endl;
};
