//----------------------------------------------------------------------------
/// @file singleton.hpp
/// @brief
///
/// @author Copyright (c) 2010 2013 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_UTIL_SINGLETON_HPP
#define __CNTREE_UTIL_SINGLETON_HPP

#include <mutex>
#include <boost/countertree/util/spinlock.hpp>
#include <cassert>

namespace boost
{
namespace countertree
{
namespace util
{

//##########################################################################
//                                                                        ##
//       ####################################################             ##
//       #                                                  #             ##
//       #                   C L A S S                      #             ##
//       #               S I N G L E T O N  < T >           #             ##
//       #                                                  #             ##
//       ####################################################             ##
//                                                                        ##
//##########################################################################

//---------------------------------------------------------------------------
/// @class singleton
///
/// @remarks This is a implementation of the Meyers singleton fo global static
///          variables
//---------------------------------------------------------------------------
template <class T >
class singleton : private T
{
private:
//***************************************************************************
//             P R I V A T E      F U N C T I O N S
//***************************************************************************
//---------------------------------------------------------------------------
//  function : constructor
/// @brief
/// @param [in]
//---------------------------------------------------------------------------
singleton() { };
//---------------------------------------------------------------------------
//  function : destructor
/// @brief
//---------------------------------------------------------------------------
~singleton(){};

public:
//---------------------------------------------------------------------------
//  function : instance
/// @brief return a reference to the unique instance of the class
/// @param [in]
/// @return reference  to the instance
//---------------------------------------------------------------------------
static T &instance()
{   static singleton <T> msgltn;
    return(msgltn);
};
//***************************************************************************
};// E N D    C L A S S    S I N G L E T O N
//***************************************************************************

//##########################################################################
//                                                                        ##
//       ####################################################             ##
//       #                                                  #             ##
//       #                   C L A S S                      #             ##
//       #         S I N G L E T O N _ C N C  < T  >        #             ##
//       #                                                  #             ##
//       ####################################################             ##
//                                                                        ##
//##########################################################################

//---------------------------------------------------------------------------
/// @class singleton_cnc
///
/// @remarks This is a implementation of the Meyers singleton
//---------------------------------------------------------------------------
template <class T >
class singleton_cnc : private T
{
private:
//***************************************************************************
//      P R I V A T E    D E F I N I T I O N S  &  V A R I A B L E S
//***************************************************************************

//***************************************************************************
//             P R I V A T E      F U N C T I O N S
//***************************************************************************
//---------------------------------------------------------------------------
//  function : constructor
/// @brief
/// @param [in]
//---------------------------------------------------------------------------
singleton_cnc() { };
//---------------------------------------------------------------------------
//  function : destructor
/// @brief
//---------------------------------------------------------------------------
~singleton_cnc(){};

public:
//---------------------------------------------------------------------------
//  function : instance
/// @brief return a reference to the unique instance of the class
/// @param [in]
/// @return reference  to the instance
//---------------------------------------------------------------------------
static T &instance()
{   spinlock  LD ;
    std::lock_guard< spinlock >  ML ( LD);
    static singleton_cnc<T> msgltn;
    return(msgltn);
};
//***************************************************************************
};// E N D    C L A S S     S I N G L E T O N _ C N C
//***************************************************************************
//
//##########################################################################
//                                                                        ##
//       ####################################################             ##
//       #                                                  #             ##
//       #                   C L A S S                      #             ##
//       #     S I N G L E T O N _ T H R E A D   < T  >     #             ##
//       #                                                  #             ##
//       ####################################################             ##
//                                                                        ##
//##########################################################################

#ifdef  _MSC_VER
    #define THREAD_LOCAL __declspec(thread )
#else
    #define THREAD_LOCAL __thread
#endif
//---------------------------------------------------------------------------
/// @class seed
///
/// @remarks This a singleton thread_local
//---------------------------------------------------------------------------
template <class T >
struct seed
{   T *P ;
    uint32_t N ;
};
/*
//---------------------------------------------------------------------------
/// @class singleton_thread
///
/// @remarks This is a singleton thread_local with the C++03, which don't
//           initialize the thread_local variables. and in many compilers use
//           the reserve word __thread instead of thread_local
//---------------------------------------------------------------------------
template <class T >
class singleton_thread
{
private :
static THREAD_LOCAL seed<T> D;

public:

singleton_thread ( void)
{   //---------------- begin ------------
    if ( D.N == 0 ) D.P = new T();
    D.N++;
}
~singleton_thread( void)
{   //------------- begin ---------------
    assert ( D.N != 0 );
    if ( D.N == 1)
    {   delete D.P ;
        D.P = NULL ;
    };
    D.N-- ;
};

T & operator() ( void ){  return  * D.P ;};

T * operator -> () {   return D.P ;};
T & operator * () { return *D.P;};

};// ------------------ end class unique_object ---------------------

template <class T >
thread_local seed<T> singleton_thread<T>::D = { NULL , 0 } ;
*/

//-------------------------------------------------------------------
// This singleton use the thread_local as defined in C++11
//-------------------------------------------------------------------
template <class T>
class singleton_thread_local : private T
{
private :
    singleton_thread_local () { };
    ~singleton_thread_local () { };
public :
    static T & instance ()
    {
        static thread_local singleton_thread_local <T> Aux ;
        return Aux ;
    }

};
//
//##########################################################################
//                                                                        ##
//       ####################################################             ##
//       #                                                  #             ##
//       #                   C L A S S                      #             ##
//       #  C O N F I G _ S I N G L E T O N < T , C N C >   #             ##
//       #                                                  #             ##
//       ####################################################             ##
//                                                                        ##
//##########################################################################
//
template <class T , bool cnc >
struct cnf_singleton ;

template <class T >
struct cnf_singleton <T , true >
{   typedef singleton_cnc<T>  data;
};
template <class T >
struct cnf_singleton <T , false >
{   typedef singleton_thread_local<T>  data;
};

//***************************************************************************
};//    End namespace util
};//    End namespace countertree
};//    End namespace boost
//***************************************************************************
#endif
