//----------------------------------------------------------------------------
/// @file parallel_sort.hpp
/// @brief Parallel Sort algorithm
///
/// @author Copyright (c) 2010 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_SORT_PARALLEL_SORT_HPP
#define __CNTREE_SORT_PARALLEL_SORT_HPP

#include <boost/countertree/util/atomic.hpp>
#include <boost/countertree/util/stack_cnc.hpp>
#include <boost/countertree/util/sort/intro_sort.hpp>
#include <boost/countertree/util/sort/indirect_sort.hpp>
#include <vector>


namespace boost
{
namespace countertree
{
namespace util
{
namespace sort
{
using namespace boost::countertree::util;


/*

template <class iter_t>
struct parallel_sort
{   //------------------------- begin ----------------------
    //typedef typename std::remove_reference<decltype(*first)>::type value_t ;
    typedef typename cntree::srt::token_level < iter_t> token_t ;
    size_type MaxPerThread = 65536;
    static const size_t Min_Parallel = 65536 ;

    //-------------------------------------------------------------------------
    //                 Variables
    //-------------------------------------------------------------------------
    cntree::util::stack_cnc <token_t> ST ;
    std::atomic<difference_type > ND ;

    //-------------------------------------------------------------------------
    //       Constructor
    //-------------------------------------------------------------------------
    parallel_sort ( iter_t first, iter_t last, uint32_t NThread = 1)
    {   //------------------------- begin ----------------------
        assert ( NThread > 0)  ;
        difference_type N = last - first;
        uint32_t Level = (cntree::util::NBits (N)) <<1 ;
        MaxPerThread = N / (NThread <<3);

        if ( N < Min_Parallel or NThread == 1 )
        {   cntree::srt::intro_sort ( first, last, Level) ;
            return ;
        } ;
        //-------------------------------------------------------------------
        //
        //-------------------------------------------------------------------
        ST.reserve ( Level * NThread );
        std::atomic_store (&ND, N);
        ST.emplace_back ( first, last, Level);

        std::vector<std::thread> VT (NThread);

        for ( uint32_t i  =0  ; i < NThread ; ++i)
            VT[i] = std::thread ( &parallel_sort::srt_thread, this );
        for ( uint32_t i  =0  ; i < NThread ; ++i)
            VT[i].join();
    };

    void sort_thread( )
    {   //------------------------- begin ----------------------
       token_t tk ;
        while ( std::atomic_load_explicit ( &ND, std::memory_order_acquire ) != 0)
        {   //--------------------------------------------------------------
            //                 read from ST
            //--------------------------------------------------------------
            if ( not ST.pop_copy_back( tk) )
            {   std::this_thread::yield() ;
                continue ;
            };
            difference_type N = tk.last - tk.first ;
            if ( N < MaxPerThread or tk.level == 0)
            {   std::atomic_fetch_sub_explicit (& ND, N ,std::memory_order_acq_rel);
                cntree::srt::intro_sort ( tk.first, tk.last, tk.level) ;
                continue ;
            };

            //----------------------------------------------------------------
            //                     split
            //----------------------------------------------------------------
            typedef typename std::remove_reference<decltype(*tk.first)>::type value_t ;
            value_t val = *(tk.first +(N>>1));

            iter_t c_first = tk.first , c_last  = tk.last-1;
            while ( *c_first < val ) ++c_first ;
            while ( *c_last > val ) --c_last ;
            while (not( c_first > c_last ))
            {   std::swap ( *(c_first++), *(c_last--));
                while ( *c_first < val ) ++c_first;
                while ( *c_last  > val ) --c_last ;
            }; // End while

            //---------------------------------------------------------------------
            //   Store the tokens en ST
            //---------------------------------------------------------------------
            token_t tk1 ( tk.first, c_last+1, tk.level-1) ;
            token_t tk2 ( c_first , tk.last,  tk.level-1) ;
            auto N1 = (tk1.last - tk1.first);
            auto N2 = (tk2.last - tk2.first);
            auto N3 = N - N1- N2 ;
            if ( N3 != 0)
            {   std::atomic_fetch_sub_explicit (& ND, N3 ,std::memory_order_acq_rel);
            };
            if ( N2 > N1 )
            {   ST.push_back ( tk2);
                ST.push_back ( tk1);
            }
            else
            {   ST.push_back ( tk1);
                ST.push_back ( tk2);
            } ;

            //std::cout<<"["<<(tk.last - tk.first)<<" "<<(tk1.last - tk1.first)<<" , "<<(tk2.last - tk2.first )<<"]\n";
        }; // end while
    } ;
};
*/


template    < class iter_t,
              typename compare = std::less < typename iter_value<iter_t>::type>
            >
struct parallel_sort_comp
{   //------------------------- begin ----------------------
    //typedef typename std::remove_reference<decltype(*first)>::type value_t ;
    typedef  token_level < iter_t> token_t ;
    size_type MaxPerThread = 65536;
    static const size_t Min_Parallel = 65536 ;
    compare comp;

    //-------------------------------------------------------------------------
    //                 Variables
    //-------------------------------------------------------------------------
    stack_cnc <token_t> ST ;
    std::atomic<difference_type > ND ;
    //std::vector<std::thread> VT ;

    //-------------------------------------------------------------------------
    //       Constructor
    //-------------------------------------------------------------------------
    parallel_sort_comp (iter_t first, iter_t last, NThread NT= NThread() ):parallel_sort_comp ( first,last, compare(),NT){} ;

    parallel_sort_comp ( iter_t first, iter_t last, compare comp1, NThread NT= NThread())
                        :comp(comp1)
    {   //------------------------- begin ----------------------
        difference_type N = last - first;
        assert ( N >=0);
        uint32_t Level = (NBits (N)) <<1 ;
        MaxPerThread = N / (NT() <<3);

        if ( (size_t)N < Min_Parallel or NT() == 1 )
        {   intro_sort ( first, last, Level,comp) ;
            return ;
        } ;
        //-------------------------------------------------------------------
        //
        //-------------------------------------------------------------------
        ST.reserve ( Level * NT() );
        std::atomic_store (&ND, N);
        ST.emplace_back ( first, last, Level);

        std::vector<std::thread> VT (NT());

        for ( uint32_t i  =0  ; i < NT() ; ++i)
            VT[i] = std::thread ( &parallel_sort_comp::sort_thread, this );
        for ( uint32_t i  =0  ; i < NT() ; ++i)
            VT[i].join();
    };

    void sort_thread( )
    {   //------------------------- begin ----------------------
        token_t tk ;
        //while ( std::atomic_load_explicit ( &ND, std::memory_order_acquire ) != 0)
        while ( atomic_read ( ND) != 0)
        {   //--------------------------------------------------------------
            //                 read from ST
            //--------------------------------------------------------------
            if ( not ST.pop_copy_back( tk) )
            {   std::this_thread::yield() ;
                continue ;
            };
            difference_type N = tk.last - tk.first ;
            if ( (size_type)N <= MaxPerThread or tk.level == 0)
            {   //std::atomic_fetch_sub_explicit (& ND, N ,std::memory_order_acq_rel);
                atomic_sub ( ND, N );
                intro_sort ( tk.first, tk.last, tk.level,comp) ;
                continue ;
            };

            //----------------------------------------------------------------
            //                     split
            //----------------------------------------------------------------
            //typedef typename std::remove_reference<decltype(*tk.first)>::type value_t ;
            typedef typename iter_value<iter_t>::type  value_t ;
            value_t val = const_cast < value_t &> (*(tk.first +(N>>1)) );

            iter_t c_first = tk.first , c_last  = tk.last-1;
            while ( comp(*c_first, val )) ++c_first ;
            while ( comp(val ,*c_last ) ) --c_last ;
            while (not( c_first > c_last ))
            {   std::swap ( *(c_first++), *(c_last--));
                while ( comp(*c_first,val) ) ++c_first;
                while ( comp(val, *c_last) ) --c_last ;
            }; // End while

            //---------------------------------------------------------------------
            //   Store the tokens en ST
            //---------------------------------------------------------------------
            token_t tk1 ( tk.first, c_last+1, tk.level-1) ;
            token_t tk2 ( c_first , tk.last,  tk.level-1) ;
            auto N1 = (tk1.last - tk1.first);
            auto N2 = (tk2.last - tk2.first);
            auto N3 = N - N1- N2 ;
            if ( N3 != 0)
            {   //std::atomic_fetch_sub_explicit (& ND, N3 ,std::memory_order_acq_rel);
                atomic_sub ( ND, N3 );
            };
            if ( N2 > N1 )
            {   ST.push_back ( tk2);
                ST.push_back ( tk1);
            }
            else
            {   ST.push_back ( tk1);
                ST.push_back ( tk2);
            } ;

            //std::cout<<"["<<(tk.last - tk.first)<<" "<<(tk1.last - tk1.first)<<" , "<<(tk2.last - tk2.first )<<"]\n";
        }; // end while
    } ;
};
//------------------------------------------------------------------------------
// These functions are for to select the correct format depending of the number
// and type of the parameters
//------------------------------------------------------------------------------
template    < class iter_t,
              typename compare = std::less < typename iter_value<iter_t>::type>
            >
void parallel_sort ( iter_t first, iter_t last , NThread NT = NThread() )
{   parallel_sort_comp <iter_t> ( first, last, NT);
};

template    < class iter_t,
              typename compare = std::less < typename iter_value<iter_t>::type>
            >
void parallel_sort ( iter_t first, iter_t last, compare comp1 , NThread NT = NThread() )
{   parallel_sort_comp<iter_t,compare> ( first, last,comp1,NT);
};

template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
void indirect_parallel_sort ( iter_t first, iter_t last , NThread NT = NThread() )
{  	parallel_indirect_sort_method<parallel_sort_comp, iter_t,compare> ( first , last, compare() );
};

/*
//---------------------------------------------------------------------------
//  template <class iter_t  >
//  void parallel_sort ( iter_t first, iter_t last,
//                       uint32_t NThread = NThread_HW );
//
//  template <class iter_t , typename compare  >
//  void parallel_sort ( iter_t first, iter_t last,
//                       compare comp1 , uint32_t NThread= NThread_HW);
//
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// This struct is for to capture the number of threads from any
// integral format passed as third parameter
// If the third paramenter is not integral it's a comparator
//----------------------------------------------------------------------------
template < class iter_t,typename compare ,bool SW> struct selector_sort ;

template <class iter_t , typename compare >
struct selector_sort<iter_t,compare,false>
{   selector_sort( iter_t first, iter_t last, compare comp1  )
    {   parallel_sort_comp<iter_t,compare> ( first,last,comp1);
    };
};

template <class iter_t , typename Aux   >
struct selector_sort<iter_t,Aux,true>
{   selector_sort ( iter_t first, iter_t last, Aux NThread )
    {   parallel_sort_comp<iter_t> (first, last,NThread);
    };
};

//------------------------------------------------------------------------------
// These functions are for to select the correct format depending of the number
// and type of the parameters
//------------------------------------------------------------------------------
template <class iter_t  >
void parallel_sort ( iter_t first, iter_t last  )
{   parallel_sort_comp <iter_t> ( first, last);
};

template <class iter_t , typename compare   >
void parallel_sort ( iter_t first, iter_t last, compare comp1 , uint32_t NThread )
{   parallel_sort_comp<iter_t,compare> ( first, last,comp1,NThread);
};

template <class iter_t , typename compare   >
void parallel_sort ( iter_t first, iter_t last, compare comp1 )
{   selector_sort<iter_t,compare,std::is_integral<compare>::value> ( first, last,comp1);
};



template <class iter_t , typename compare = std::less < typename iter_value<iter_t>::type > >
void parallel_sort ( iter_t first, iter_t last, compare comp1, uint32_t NThread =NThread_HW)
{   //----------------------------begin-----------------------------------------------------
    parallel_sort_comp<iter_t,compare>( first, last, comp1, NThread );
};
template <class iter_t >
void parallel_sort ( iter_t first, iter_t last, uint32_t NThread =NThread_HW)
{   //----------------------------begin-----------------------------------------------------
    typedef std::less < typename iter_value<iter_t>::type >  compare ;
    parallel_sort_comp<iter_t,compare>( first, last, compare(), NThread );
};
*/
//
//****************************************************************************
};//    End namespace sort
};//    End namespace util
};//    End namespace countertree
};//    End namespace boost
//****************************************************************************
//
#endif


